#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
OMEGAlpes installation script

:authors: B. DELINCHANT, S. HODENCQ, Y. MARECHAL, L. MORRIET,
          C. PAJOT, F. WURTZ
:license: Apache Software License 2.0
:version: 0.3.1
"""

from setuptools import setup, find_packages

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (0, 3, 1)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


setup(

    name='omegalpes',
    version=__version__,
    packages=["omegalpes",
              "omegalpes.energy",
              "omegalpes.energy.io",
              "omegalpes.energy.units",
              "omegalpes.general",
              "omegalpes.general.optimisation",
              "omegalpes.general.utils",
              "omegalpes.actor",
              "omegalpes.actor.operator_actors",
              "omegalpes.actor.regulator_actors",
              "omegalpes.energy.buildings"
              ],
    author="B. DELINCHANT, S. HODENCQ, Y. MARECHAL, L. MORRIET, "
           "C. PAJOT, V. REINBOLD, F. WURTZ",
    author_email='omegalpes-users@groupes.renater.fr',
    description="OMEGAlpes is a linear energy systems modelling library",
    long_description_content_type='text/markdown',
    long_description=open('README.md').read(),
    install_requires=[
        "PuLP >= 1.6.10",
        "Matplotlib >= 2.2.2",
        "Numpy >= 1.14.2",
        "Pandas >= 0.25.0"
    ],
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering",
    ],

)



#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
**This module creates the Time object**

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import numpy as np
import pandas as pd

__docformat__ = "restructuredtext en"


class TimeUnit:
    """
    **Description**

        Class defining the studied time period.

    **Attributes**

     - DATES : dated list of simulation steps
     - DT : delta t between values in hours (int or float), i.e. 1/6 will
       be 10 minutes.
     - LEN : number of simulation steps (length of DATES)
     - I : index of time ([0 : LEN])

    """

    def __init__(self, start='01/01/2018', end=None, periods=None, dt=1,
                 verbose=True):
        """
        :param start: start of the studied time period. European or DateTime
            format
        :param end: end of the studied time period. European or DateTime format
        :param periods: number of time step between the start and the end of
            the studied time period
        :param dt: time step size i.e. 1 being 1 hour,  1/6 will be 10 minutes.
        :param verbose: printing the studied time period if True
        """
        start = convert_european_format(start)
        end = convert_european_format(end)

        self.DATES = pd.date_range(start=start, end=end, periods=periods,
                                   freq=str(dt) + 'H')
        self.DT = dt
        self.LEN = len(self.DATES)
        self.I = np.arange(self.LEN)
        if verbose:
            self.print_studied_period()

    @property
    def get_days(self):
        """
        Getting days for the studied period

        :return all_days: list of days of the studied period
        """
        all_days = []  # Initialize the list of days

        for date in self.DATES:
            day = date.date()  # Get dates without the hour

            if day not in all_days:
                all_days.append(day)  # Gather all days in a list

        return all_days

    def get_date_for_index(self, index):
        """
        Getting a date for a given index

        :param index: int value for the index of the wanted dated, between 0
            and LEN (it must be in the studied period)
        """
        if isinstance(index, int):
            if index < self.LEN:
                return self.DATES[index]
            else:
                raise ValueError("The index value should be lower than the "
                                 "number of periods")
        else:
            raise TypeError("You should enter an integer value for the index "
                            "in order to get the associated date.")

    def get_index_for_date(self, date='YYYY-MM-DD HH:MM:SS'):
        """
        Getting the index associated with a date

        :param date: date the index of is wanted. Format YYYY-MM-DD
            HH:MM:SS, must be within the studied period and consistent with the
            timestep value
        """
        date = convert_european_format(date)

        date_tstp = pd.Timestamp(date)

        return list(self.DATES).index(date_tstp)

    def get_index_for_date_range(self, starting_date='YYYY-MM-DD HH:MM:SS',
                                 end=None, periods=None):
        """
        Getting a list of index for a date range

        :param starting_date: starting date of the wanted index
        :param end: ending date of the wanted index
        :param periods: number of periods from the starting_date of the
            wanted index
        :return index_list: list of indexes for the given dates
        """
        starting_date = convert_european_format(starting_date)
        index_list = []
        for date in pd.date_range(start=starting_date, end=end, periods=periods,
                                  freq=str(self.DT) + 'H'):
            index_list.append(self.get_index_for_date(date))

        return index_list

    def print_studied_period(self):
        start_date = self.get_date_for_index(0)
        end_date = self.get_date_for_index(-1)
        print(('You are studying the period from {0} to {1}'.format(start_date,
                                                                    end_date)))

    def get_working_days(self, country='France'):
        exec('from workalendar.europe import ' + country)
        cal = eval(country + '()')

        working_days = [day for day in self.get_days if cal.is_working_day(day)]

        return working_days

    def get_working_dates(self, month_range=range(12), hour_range=range(0, 24),
                          country='France'):
        exec('from workalendar.europe import ' + country)
        cal = eval(country + '()')
        working_dates = [date for date in self.DATES
                         if cal.is_working_day(date)
                         if date.hour in hour_range
                         if date.month in month_range]

        return working_dates

    def get_non_working_days(self, country='France'):
        exec('from workalendar.europe import ' + country)
        cal = eval(country + '()')

        working_days = [day for day in self.get_days
                        if not cal.is_working_day(day)]

        return working_days

    def get_non_working_dates(self, month_range=range(12), hour_range=range(24),
                              country='France'):
        exec('from workalendar.europe import ' + country)
        cal = eval(country + '()')
        working_dates = [date for date in self.DATES
                         if not cal.is_working_day(date)
                         if date.hour in hour_range
                         if date.month in month_range]

        return working_dates


def convert_european_format(date):
    """
        Converting a date with an european format DD/MM/YYYY
        into a datetime format YYYY-MM-DD or return

    :param date: date in european format
    :return: date in format datetime
    """

    try:
        # If date format begins with 'XX/XX' consider european date format
        if date[2] == '/':
            new_date = pd.to_datetime(date, dayfirst=True)
        else:
            new_date = date

    except TypeError:
        new_date = date

    return new_date



#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
**This module defines the energy nodes that will allow energy transmission
between the various energy units and conversion units**

The energy_node module includes the EnergyNode class for energy transmission
between production, consumption, conversion and storage. Defining several
energy nodes and exporting/importing energy between them can also allow for a
better demarcation of the energy system.

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from pulp import LpBinary
from ..general.optimisation.core import OptObject
from ..energy.io.poles import FlowPole, Epole
from .units.energy_units import EnergyUnit, AssemblyUnit
from ..general.optimisation.elements import *

__docformat__ = "restructuredtext en"


class EnergyNode(OptObject):
    """
    This class defines an energy node.
    """

    def __init__(self, time, name, energy_type=None, operator=None):
        OptObject.__init__(self, name=name, description='Energy Node')

        self.time = time
        self.energy_type = energy_type
        self._connected_energy_units = []
        self._poles_list = []
        self._exports = []
        self._imports = []
        self.operator = operator

    @property
    def get_connected_energy_units(self):
        """ Return the list of connected EnergyUnits in the EnergyNode """
        return self._connected_energy_units

    @property
    def get_flows(self):
        """ Get all the power flows of the energy node
        :rtype: list
        :return: list of power flows
        """
        return [pole['p'] for pole in self.get_poles]

    @property
    def get_poles(self):
        """ Return the list of energy poles in the EnergyNode """
        return self._poles_list

    @property
    def get_exports(self):
        """ Return the list of exports to the EnergyNode"""
        return self._exports

    @property
    def get_imports(self):
        """ Return the list of imports to the EnergyNode"""
        return self._imports

    def add_connected_energy_unit(self, unit):
        """ Add an EnergyUnit to the connected_units list """
        if isinstance(unit, EnergyUnit):
            if unit not in self.get_connected_energy_units:
                self._connected_energy_units.append(unit)

                if unit.energy_type is None:
                    unit.energy_type = self.energy_type

                elif self.energy_type is None:
                    self.energy_type = unit.energy_type

                elif unit.energy_type != self.energy_type:
                    raise TypeError('You cannot connect an {0} EnergyUnit to a'
                                    ' {1} EnergyNode.'.format(unit.energy_type,
                                                              self.energy_type)
                                    )
                self.add_pole(unit.poles[1])
        elif isinstance(unit, AssemblyUnit):
            raise TypeError(
                "The unit {0} is an AssemblyUnit: you should connect "
                "the Energy units within this assembly unit to the "
                "node, not the whole unit!".format(unit.name))
        else:
            raise TypeError(
                "The unit {0} to connect to an EnergyNode should be an "
                "EnergyUnit and is {1}".format(unit, type(unit)))

    def add_pole(self, pole: Epole) -> None:
        """ Add an energy pole to the poles_list

        :param pole: Epole
        """
        assert isinstance(pole, Epole)
        if pole not in self.get_poles:
            self._poles_list.append(pole)

    def connect_units(self, *units):
        """
        Connecting all EnergyUnit to the EnergyNode

        :param units: EnergyUnits connected to the EnergyNode
        :type units: EnergyUnit
        """

        for unit in units:
            self.add_connected_energy_unit(unit)

    def set_power_balance(self):
        """ Set the power balance equation for the EnergyNode """
        exp_t = ''

        for pole in self.get_poles:
            if not isinstance(pole, FlowPole):
                raise TypeError('Wrong Type of arguments,\
                arguments should be an FlowPole')
            if isinstance(pole[pole.flow], Quantity) and (pole.flow in pole):
                if pole[pole.direction] == 'in':
                    exp_t += '-' + pole[pole.flow].parent.name + \
                             '_' + pole[pole.flow].name + '[t]'
                elif pole[pole.direction] == 'out':
                    exp_t += '+' + pole[pole.flow].parent.name + \
                             '_' + pole[pole.flow].name + '[t]'
        exp_t += ' == 0'

        if exp_t[0] == '+':
            exp_t = exp_t[1:]

        setattr(self, 'power_balance', DefinitionDynamicConstraint(exp_t=exp_t,
                                                                   name='power_balance',
                                                                   parent=self))

    def create_export(self, node, export_min, export_max):
        """Create the export from the EnergyNode (self)
        to the EnergyNode (node)

        :param node: EnergyNode to whom power can be exported
        :param export_min: Minimal value of exported power when there is export
        :param export_max: Maximal value of exported power when there is export
        :return: Quantity that defines the power exported
        """
        energy_export = Quantity(
            name='energy_export_to_{0}'.format(node.name), opt=True, lb=0,
            ub=export_max, vlen=self.time.LEN, parent=self)

        is_exporting = Quantity(
            name='is_exporting_to_{0}'.format(node.name),
            description='The node is exporting :1 or not :0',
            vtype=LpBinary, vlen=self.time.LEN, parent=self)

        setattr(self, 'energy_export_to_{0}'.format(node.name),
                energy_export)
        setattr(self, 'is_exporting_to_{0}'.format(node.name), is_exporting)

        if isinstance(export_min, (int, float)):
            set_export_min = TechnicalDynamicConstraint(
                exp_t='{0}_is_exporting_to_{1}[t] * {2} <= '
                      '{0}_energy_export_to_{1}[t]'.format(self.name,
                                                           node.name,
                                                           export_min),
                name='set_export_min_{0}_min'.format(node.name), parent=self)
            setattr(self, 'set_export_to_{0}_min'.format(node.name),
                    set_export_min)

        elif isinstance(export_min, list):
            set_export_min = TechnicalDynamicConstraint(
                exp_t='{0}_is_exporting_to_{1}[t] * {2}[t] <= '
                      '{0}_energy_export_to_{1}[t]'.format(self.name,
                                                           node.name,
                                                           export_min),
                name='set_export_min', parent=self)
            setattr(self, 'set_export_min', set_export_min)

        if isinstance(export_max, (int, float)):
            set_export_max = TechnicalDynamicConstraint(
                exp_t='{0}_is_exporting_to_{1}[t] * {2} >= '
                      '{0}_energy_export_to_{1}[t]'.format(self.name,
                                                           node.name,
                                                           export_max),
                name='set_export_to_{0}_max'.format(node.name), parent=self)
            setattr(self, 'set_export_to_{0}_max'.format(node.name),
                    set_export_max)
        elif isinstance(export_max, list):
            set_export_max = TechnicalDynamicConstraint(
                exp_t='{0}_is_exporting_to_{1}[t] * {2}[t] >= '
                      '{0}_energy_export_to_{1}[t]'.format(self.name,
                                                           node.name,
                                                           export_max),
                name='set_export_max', parent=self)
            setattr(self, 'set_export_max', set_export_max)

        self._exports.append(energy_export)
        node._imports.append(energy_export)

        return energy_export

    def export_to_node(self, node, export_min=1e-5, export_max=1e+5):
        """ Add an export of power from the node to another node

        :param node: EnergyNode to whom power can be exported
        :param export_min: Minimal value of exported power when there is export
        :param export_max: Maximal value of exported power when there is export
        """
        if not isinstance(node, EnergyNode):
            raise TypeError(
                'The node {0} should be an EnergyNode'.format(node))

        if node.energy_type != self.energy_type:
            raise AttributeError(
                'You cannot export energy from an EnergyNode '
                'with energy_type "{0}" to an EnergyNode with '
                'energy_type "{1}"'.format(node.energy_type, self.energy_type))
        else:
            energy_export = self.create_export(node, export_min, export_max)

            self.add_pole(Epole(p=energy_export, direction='in',
                                energy_type=self.energy_type))
            node.add_pole(Epole(p=energy_export, direction='out',
                                energy_type=self.energy_type))

    def import_from_node(self, node, import_min=1e-5, import_max=1e5):
        """

        :param node: EnergyNode from whom power can be imported
        :param import_min: Minimal value of imported power when there is import
        :param import_max: Maximal value of imported power when there is import
        """

        node.export_to_node(self, export_min=import_min, export_max=import_max)

    def is_import_flow(self, flow):
        """ Get if the power flow is an import or not"""
        is_import = flow in self.get_imports
        return is_import

    def is_export_flow(self, flow):
        """ Get if the power flow is an export or not"""
        is_export = flow in self.get_exports
        return is_export



#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
**This module includes the following maths utils**

 It contains the following methods:
    - def_abs_value() : Define easily absolute value for quantity

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from pulp import LpBinary
from ..optimisation.elements import *

__docformat__ = "restructuredtext en"


def def_abs_value(quantity, q_min, q_max):
    """

    :param quantity: Quantity whose absolute value is wanted
    :param q_min: Minimal value of the quantity (negative value)
    :param q_max: Maximal value of the quantity (positive value)

    :return: A new Quantity whose values equal absolute values of the
        initial Quantity
    """

    if q_max <= 0:
        raise ValueError('If q_max <= 0, the absolute value of your quantity '
                         'is -1 * quantity.')
    if q_min >= 0:
        raise ValueError('If q_min >= 0, the absolute value of your quantity '
                         'is your quantity.')

    q_name = getattr(quantity, 'name')
    q_len = getattr(quantity, 'vlen')
    parent = getattr(quantity, 'parent')

    abs_value = Quantity(name='{}_abs'.format(q_name), opt=True, lb=0,
                         ub=q_max, vlen=q_len, parent=parent)
    value_pos = Quantity(name='{}_pos'.format(q_name), opt=True, lb=0, ub=q_max,
                         vlen=q_len, parent=parent)
    value_neg = Quantity(name='{}_neg'.format(q_name), opt=True, lb=0,
                         ub=-q_min, vlen=q_len, parent=parent)
    is_pos = Quantity(name='is_{}_pos'.format(q_name), opt=True, vtype=LpBinary,
                      vlen=q_len, parent=parent)

    setattr(parent, '{}_abs'.format(q_name), abs_value)
    setattr(parent, '{}_pos'.format(q_name), value_pos)
    setattr(parent, '{}_neg'.format(q_name), value_neg)
    setattr(parent, 'is_{}_pos'.format(q_name), is_pos)

    if q_len == 1:
        set_decomp = DefinitionConstraint(
            exp='{0}_{1} == {0}_{1}_pos - {0}_{1}_neg'.format(parent.name,
                                                              q_name),
            name='set_{}_decomp'.format(q_name), parent=parent)

        set_if_pos = DefinitionConstraint(
            exp='{0}_{1}_pos <= {2} * {0}_is_{1}_pos'.format(parent.name,
                                                             q_name, q_max),
            name='set_if_{}_pos'.format(q_name), parent=parent)

        set_if_neg = DefinitionConstraint(
            exp='{0}_{1}_neg <= {2} * ({0}_is_{1}_pos - 1)'.format(parent.name,
                                                                   q_name,
                                                                   q_min),
            name='set_if_{}_neg'.format(q_name), parent=parent)
        set_abs = DefinitionConstraint(
            exp='{0}_{1}_abs == {0}_{1}_pos + {0}_{1}_neg'.format(parent.name,
                                                                  q_name),
            name='set_{}_abs'.format(q_name), parent=parent)

    elif q_len > 1:
        set_decomp = DefinitionDynamicConstraint(
            exp_t='{0}_{1}[t] == {0}_{1}_pos[t] - {0}_{1}_neg[t]'.format(
                parent.name, q_name),
            name='set_{}_decomp'.format(q_name), parent=parent)

        set_if_pos = DefinitionDynamicConstraint(
            exp_t='{0}_{1}_pos[t] <= {2} * {0}_is_{1}_pos[t]'.format(
                parent.name, q_name, q_max),
            name='set_if_{}_pos'.format(q_name), parent=parent)

        set_if_neg = DefinitionDynamicConstraint(
            exp_t='{0}_{1}_neg[t] <= {2} * ({0}_is_{1}_pos[t] - 1)'.format(
                parent.name, q_name, q_min),
            name='set_if_{}_neg'.format(q_name), parent=parent)

        set_abs = DefinitionDynamicConstraint(
            exp_t='{0}_{1}_abs[t] == {0}_{1}_pos[t] + {0}_{1}_neg[t]'.format(
                parent.name, q_name),
            name='set_{}_abs'.format(q_name), parent=parent)

    setattr(parent, 'set_{}_decomp'.format(q_name), set_decomp)
    setattr(parent, 'set_if_{}_pos'.format(q_name), set_if_pos)
    setattr(parent, 'set_if_{}_neg'.format(q_name), set_if_neg)
    setattr(parent, 'set_{}_abs'.format(q_name), set_abs)

    return abs_value



#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
**This module describes the Producer actor**

 Few objectives and constraints are available.

 Objectives :
    - maximize_production
    - minimize_production
    - minimize_time_of_use
    - minimize_co2_emissions
    - minimize_costs
    - minimize_operating_cost
    - minimize_starting_cost

 Constraints :
    - energy_production_minimum
    - energy_production_maximum
    - power_production_minimum
    - power_production_maximum

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from ..operator_actors.operator_actors import OperatorActor
from ...general.optimisation.elements import *
from ...energy.units.production_units import ProductionUnit
from ...energy.energy_nodes import EnergyNode

__docformat__ = "restructuredtext en"


class Producer(OperatorActor):
    """
    **Description**

        Producer class inherits from the the class OperatorActor. It enables
        one to model an energy producer actor.
    """

    def __init__(self, name, operated_unit_list, operated_node_list=None,
                 verbose=True):
        OperatorActor.__init__(self, name=name,
                               operated_unit_type_tuple=(ProductionUnit,
                                                         EnergyNode),
                               operated_unit_list=operated_unit_list,
                               operated_node_list=operated_node_list,
                               verbose=verbose)

    # OBJECTIVES #
    # Energy objectives
    def maximize_production(self, obj_operated_unit_list=None, weight=1):
        """
        To create the objective in order to maximize the production of the
        producer's units (all or part of them).

        :param obj_operated_unit_list: List of units on which the
            objective will be applied. Might be empty.
        :param weight: Weight coefficient for the objective
        """
        final_operated_unit_list = self._check_operated_unit_list(
            obj_operated_unit_list)
        for operated_unit in final_operated_unit_list:
            operated_unit.maximize_production(weight=weight)

    def minimize_production(self, obj_operated_unit_list=None, weight=1):
        """
        To create the objective in order to minimize the production of the
        producer's units (all or part of them).

        :param obj_operated_unit_list: List of units on which the
            objective will be applied. Might be empty.
        :param weight: Weight coefficient for the objective
        """
        final_operated_unit_list = self._check_operated_unit_list(
            obj_operated_unit_list)
        for operated_unit in final_operated_unit_list:
            operated_unit.minimize_production(weight=weight)

    def minimize_time_of_use(self, obj_operated_unit_list=None, weight=1):
        """
        To create the objective in order to minimize the time of use of the
        producer's units (all or part of them).

        :param obj_operated_unit_list: List of units on which the
            objective will be applied. Might be empty.
        :param weight: Weight coefficient for the objective
        """
        final_operated_unit_list = self._check_operated_unit_list(
            obj_operated_unit_list)
        for operated_unit in final_operated_unit_list:
            operated_unit.minimize_time_of_use(weight=weight)

    # CO2 objectives
    def minimize_co2_emissions(self, obj_operated_unit_list=None, weight=1):
        """
        To create the objective in order to minimize the co2 emissions of the
        producer's units (all or part of them).
        based on the quantity "co2_emission"

        :param obj_operated_unit_list: list of the operated energy units on
            which the objective will be applied
        :param weight: weight of the objective
        """
        final_operated_unit_list = self._check_operated_unit_list(
            obj_operated_unit_list)
        for operated_unit in final_operated_unit_list:
            operated_unit.minimize_co2_emissions(weight=weight)

    # Economic objectives
    def minimize_costs(self, obj_operated_unit_list=None, weight=1):
        """
        To create the objective in order to minimize the cost of the
        producer's units (all or part of them).

        :param obj_operated_unit_list: list of the operated energy units on
            which the objective will be applied
        :param weight: weight of the objective
        """
        final_operated_unit_list = self._check_operated_unit_list(
            obj_operated_unit_list)
        for operated_unit in final_operated_unit_list:
            operated_unit.minimize_costs(weight=weight)

    def minimize_operating_cost(self, obj_operated_unit_list=None, weight=1):
        """
        To create the objective in order to minimize the operating costs of the
        producer's units (all or part of them).

        :param obj_operated_unit_list: list of the operated energy units on
            which the objective will be applied
        :param weight: weight of the objective
        """
        final_operated_unit_list = self._check_operated_unit_list(
            obj_operated_unit_list)
        for operated_unit in final_operated_unit_list:
            operated_unit.minimize_operating_cost(weight=weight)

    def minimize_starting_cost(self, obj_operated_unit_list=None, weight=1):
        """
        To create the objective in order to minimize the starting costs of the
        producer's units (all or part of them).

        :param obj_operated_unit_list: list of the operated energy units on
            which the objective will be applied
        :param weight: weight of the objective
        """
        final_operated_unit_list = self._check_operated_unit_list(
            obj_operated_unit_list)
        for operated_unit in final_operated_unit_list:
            operated_unit.minimize_starting_cost(weight=weight)

    # CONSTRAINTS #
    # Energy constraints
    def energy_production_minimum(self, min_e_tot,
                                  cst_operated_unit_list=None):
        """
        To create the actor constraint of a minimum of energy production.

        :param min_e_tot: Minimum of the total energy production over the
            period of the study
        :param cst_operated_unit_list: List of units on which the constraint
            will be applied. Might be empty.

        """
        if isinstance(min_e_tot, (int, float)):
            final_operated_unit_list = self._check_operated_unit_list(
                cst_operated_unit_list)

            production_name_string = ''
            for production_unit in final_operated_unit_list:
                production_name_string += '_{}'.format(
                    production_unit.name)

                production_p_string = ''
            for production_unit in final_operated_unit_list[:-1]:
                production_p_string += '{}_p[t] + '.format(production_unit.name)
            production_p_string += '{}_p[t]'.format(
                final_operated_unit_list[-1].name)

            cst_name = 'min_energy_prod{}'.format(production_name_string)
            exp = 'lpSum({0} for t in time.I) >= {1}'.format(
                production_p_string, min_e_tot)

            setattr(self, cst_name, ActorConstraint(exp=exp,
                                                    name=cst_name,
                                                    parent=self))

        else:
            raise TypeError('min_e_tot in energy_production_minimum should be '
                            'an int or a float')

    def energy_production_maximum(self, max_e_tot,
                                  cst_operated_unit_list=None):
        """
        To create the actor constraint of a maximum of energy production.

        :param max_e_tot: Maximum of the total energy production over the
            period of the study
        :param cst_operated_unit_list: List of units on which the constraint
            will be applied. Might be empty.

        """
        if isinstance(max_e_tot, (int, float)):
            final_operated_unit_list = self._check_operated_unit_list(
                cst_operated_unit_list)

            production_name_string = ''
            for production_unit in final_operated_unit_list:
                production_name_string += '_{}'.format(
                    production_unit.name)

            production_p_string = ''
            for production_unit in final_operated_unit_list[:-1]:
                production_p_string += '{}_p[t] + '.format(production_unit.name)
            production_p_string += '{}_p[t] '.format(
                final_operated_unit_list[-1].name)

            cst_name = 'max_energy_prod{}'.format(production_name_string)
            exp = 'lpSum({0} for t in time.I) <= {1}'.format(
                production_p_string, max_e_tot)

            setattr(self, cst_name, ActorConstraint(exp=exp,
                                                    name=cst_name,
                                                    parent=self))

        else:
            raise TypeError('max_e_tot in energy_production_maximum should be '
                            'an int or a float')

    def power_production_minimum(self, min_p, time,
                                 cst_operated_unit_list=None):
        """
        To create the constraint of a minimum of power production.

        :param min_p: Minimum of the power production. May be an int,
            float or a list with the size of the period study
        :param time: period of the study
        :param cst_operated_unit_list: List of units on which the constraint
            will be applied. Might be empty.

        """
        final_operated_unit_list = self._check_operated_unit_list(
            cst_operated_unit_list)

        production_name_string = ''
        for production_unit in final_operated_unit_list:
            production_name_string += '_{}'.format(
                production_unit.name)

        production_p_string = ''
        for production_unit in final_operated_unit_list[:-1]:
            production_p_string += '{}_p[t] + '.format(production_unit.name)
        production_p_string += '{}_p[t]'.format(
            final_operated_unit_list[-1].name)

        cst_name = 'min_power_prod{}'.format(production_name_string)

        if isinstance(min_p, (int, float)):
            exp = '{0} >= {1}'.format(production_p_string, min_p)
            setattr(self, cst_name, ActorConstraint(exp=exp,
                                                    name=cst_name,
                                                    parent=self))
        elif isinstance(min_p, list):
            if len(min_p) != time.LEN:
                raise IndexError(
                    "Your minimal power should be the size of the time "
                    "period : {} but equals {}.".format(time.LEN,
                                                        len(min_p)))
            else:
                exp_t = '{0} >= {1}[t]'.format(production_p_string, min_p)
                setattr(self, cst_name, ActorDynamicConstraint(exp_t=exp_t,
                                                               name=cst_name,
                                                               t_range='for '
                                                                       't in '
                                                                 'time.I',
                                                               parent=self))
        else:
            raise TypeError('Your minimal power should be an int, '
                            'a float or a list.')

    def power_production_maximum(self, max_p, time,
                                 cst_operated_unit_list=None):
        """
        To create the actor constraint of a maximum of power production.

        :param max_p: Minimum of the power production. May be an int,
            float or a list with the size of the period study
        :param time: period of the study
        :param cst_operated_unit_list: List of units on which the constraint
            will be applied. Might be empty.

        """
        final_operated_unit_list = self._check_operated_unit_list(
            cst_operated_unit_list)

        production_name_string = ''
        for production_unit in final_operated_unit_list:
            production_name_string += '_{}'.format(
                production_unit.name)

        production_p_string = ''
        for production_unit in final_operated_unit_list[:-1]:
            production_p_string += '{}_p[t] + '.format(production_unit.name)
        production_p_string += '{}_p[t]'.format(
            final_operated_unit_list[-1].name)

        cst_name = 'max_power_prod{}'.format(production_name_string)

        if isinstance(max_p, (int, float)):
            exp = '{0} <= {1}'.format(production_p_string, max_p)
            setattr(self, cst_name, ActorConstraint(exp=exp,
                                                    name=cst_name,
                                                    parent=self))

        elif isinstance(max_p, list):
            if len(max_p) != time.LEN:
                raise IndexError(
                    "Your maximal power should be the size of the time "
                    "period : {} but equals {}.".format(time.LEN,
                                                        len(max_p)))
            else:
                exp_t = '{0} <= {1}[t]'.format(production_p_string, max_p)
                setattr(self, cst_name, ActorDynamicConstraint(exp_t=exp_t,
                                                               name=cst_name,
                                                               t_range='for t in '
                                                                 'time.I',
                                                               parent=self))
        else:
            raise TypeError('Your maximal power should be an int, '
                            'a float or a list.')



#! usr/bin/env python3
#  -*- coding: utf-8 -*-


"""
**This module defines the operator_actor and its scope of responsibility**

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from ..actor import Actor


class OperatorActor(Actor):
    """
    **Description**

        OperatorActor class inherits from the the basic class Actor. It
        enables one to model an actor who operates energy units which is part
        of its scope of responsibility.
        An operator actor has objectives and constraints which are
        linked to the energy units he operates.

    **Attributes**

     - name : name of the actor
     - operated_unit_list:  list of the energy units operated by the actor or
       more precisely in its scope of responsibility

    """

    def __init__(self, name, operated_unit_type_tuple,
                 operated_unit_list=None, operated_node_list=None,
                 verbose=True):
        Actor.__init__(self, name=name, verbose=verbose)

        self.description = 'Operator Actor OptObject'
        self.operated_unit_list = operated_unit_list or []
        self.operated_node_list = operated_node_list or []

        for unit in self.operated_unit_list:
            self._check_unit_type(unit, operated_unit_type_tuple)
            unit._set_optobject_as_attr(optobject=self,
                                        attribute_name='operator')

    def _check_operated_unit_list(self, obj_operated_unit_list):
        """
        Check if the objective must be applied to the whole energy units
        under the scope of responsibility (if obj_operated_unit_list is
        empty); or to selected units and in this case, if the units are under
        the scope of responsibility

        :param obj_operated_unit_list: List of units on which the
            objective will be applied. Might be empty.
        :return: the list of the operated energy unit in order to apply
            the objective on these units
        """
        final_operated_unit_list = []
        if not obj_operated_unit_list:
            final_operated_unit_list = self.operated_unit_list
        else:
            if not isinstance(obj_operated_unit_list, list):
                raise TypeError("The operated unit list in order to add an "
                                "objective should be a list")
            else:
                for operated_unit in obj_operated_unit_list:
                    if operated_unit not in self.operated_unit_list \
                            and operated_unit.parent not in \
                            self.operated_node_list:
                        raise ValueError("the unit {0} must be "
                                         "operated by {1}"
                                         .format(operated_unit.name, self.name))
                    else:
                        final_operated_unit_list = obj_operated_unit_list
        return final_operated_unit_list

    def _check_unit_type(self, operated_unit, operated_unit_type_tuple):
        """
        Check if the unit type corresponds to the actor energy unit type
        otherwise it raises a TypeError

        :param operated_unit: operated energy unit on the actor scope of
            responsibility
        :param operated_unit_type_tuple: available type of the energy units
            considering the actor plus energynode type
        """
        if not issubclass(operated_unit.__class__, operated_unit_type_tuple):
            raise TypeError("The operated unit type {0} does not correspond "
                            "to the actor energy unit type {1} ".format(
                operated_unit.__class__, operated_unit_type_tuple))
